% rn mode -- an rn lookalike macro package.  % -*- mode: slang; mode: fold -*-
% This makes slrn look and feel more like rn.
% 
% This is a work in progress.  Don't depend on this remaining precisely the
% same by the time this macro set reaches version 1.0.
%
% Author: J.B. Nicholson-Owens
% License: GNU GPL
% Version: 0.11
% Copyright 2002  J.B. Nicholson-Owens
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

implements ("rn");

private variable
  % The preferences for this macro set.
  prefs = Assoc_Type[String_Type],
  preference_check = Assoc_Type[Ref_Type],

  % Subject search sort order--boolean.  Set when ^N is used,
  % unset otherwise.
  subject_search_sort_order = 0,

  % This is a bit set to indicate a range is possibly starting.
  is_a_range_request = 0,

  % The set of articles the user selected to be marked read on
  % exiting the group.  We set the first entry to be the empty
  % string for two reasons: to set up a String_Type array, and
  % to make sure where() always returns a positive integer.
  delay_mark_current_article_unread_set = [""],
  
  % The command map maps a :command syntax letter to a function
  % that implements that command.  For instance "m" is the
  % command to mark an article read.  This means the command_map
  % maps "m" to &command_mark_as_unread().
  %
  % None of the command_*() functions move either article mode or
  % group mode cursors.
  command_map = Assoc_Type[Ref_Type];

private define header_cursor_pos () %{{{
{
   % Bogus placeholder -- remove for release
   return 5;
}

%}}}
private define check_options_preference (new_setting) %{{{
{
   % Check the user's options preference setting and correct if necessary.
   variable
     % Return codes for how the check came out.
     OK = 0, BAD = 1;

   return OK;
}

%}}}
private define jump_to_article (article_index_number) %{{{
{
   % Place the cursor on an article given its index number.

   if (orelse
         { typeof (article_index_number) != Integer_Type }
         { article_index_number < 1 }
         { article_index_number > SCREEN_HEIGHT - 3 })
     {
        % Can't move the cursor to any of these locations, so indicate failure.
        return (0);
     }

   variable current_header_cursor_position = header_cursor_pos ();
   if (current_header_cursor_position < article_index_number)
     {
        () = header_down (article_index_number - current_header_cursor_position);
     }
   else if (current_header_cursor_position > article_index_number)
     {
        () = header_up (current_header_cursor_position - article_index_number);
     }
   return (1);
}

%}}}

prefs["options"] = "";
preference_check["options"] = &check_options_preference;

static define set_preference (preference, value) %{{{
{
   preference = string (preference);
   !if (assoc_key_exists (prefs, preference))
     error ("Preference does not exist: " + preference);

   variable desired_type = typeof (prefs[preference]);
   if (andelse
         { desired_type != Null_Type }
         { typeof (value) != desired_type })
     {
        verror ("Wrong type for %s: This preference wants %S not %S",
                preference, desired_type, typeof (value));
     }
   if (andelse
        { assoc_key_exists (preference_check, preference) }
        { @preference_check[preference] (value) })
     {
        vmessage ("Warning: Invalid setting for %s -- using %S instead",
                  preference, prefs[preference]);
        return ();
     }

   prefs[preference] = value;
}
%}}}

% Group mode commands

static define group_mode_prompt () %{{{
{
   message (sprintf ("******** %d unread articles in %s--read now? [ynq]",
                         group_unread (), current_newsgroup ()));
}

%}}}
static define article_mode_prompt () %{{{
{
   message ("End of article--what next? [npq]");
}

%}}}
static define group_mode_hook () %{{{
{
   % Leave cursor at a group with unread articles (scan down for next
   % such group).  If there are no groups with unread articles, quit.
   variable starting_newsgroup = current_newsgroup ();
   if (starting_newsgroup == "") call ("quit");

   while (group_unread () == 0)
     {
        !if (group_down_n (1))
          {
             % Leave group mode cursor on the starting_newsgroup and return.
             if (is_group_mode ())
               {
                  % We can be quick about it!
                  set_input_string (starting_newsgroup);
                  call ("add_group");
               }
             else
               {
                  % We have to be slow about it :-(
                  while (group_up_n (10000));    % skip to the top quickly.
                  while (current_newsgroup () != starting_newsgroup)
                    {
                       !if (group_down_n (1))
                         {
                            return ();           % unexpected error here!
                         }
                    }
               }
             return ();
          }
     }

   % Prompt the user.
   group_mode_prompt ();
}
() = register_hook ("group_mode_hook", "rn->group_mode_hook");
%}}}
static define enter_article_mode () %{{{
{
   % y or space -- enter article mode.
   %
   % All the groups are read.  What does rn do here?
   % I don't know, so we'll call it quits.

   do
     {
        !if (select_group ()) return ();
        else _clear_error ();
     }
   while (group_down_n (1));
}
definekey ("rn->enter_article_mode", "y", "group");
definekey ("rn->enter_article_mode", " ", "group");

%}}}
static define enter_article_mode_then_do_something_else () %{{{
{
   % .command -- enter article mode (as above) then do command in article mode
   variable command = read_mini (sprintf ("******** %d unread articles in %s--read now? [ynq] .",
                                          group_unread (), current_newsgroup ()),
                                 "", "");

   enter_article_mode ();

   % Push the command to complete in article mode. (article_mode_hook() and
   % all the other article mode level hooks will have executed before any
   % of this executes).
   !if (is_group_mode ())
     {
        variable i = 0;
        for (i = strlen (command) - 1; i != -1; i--)
          {
             ungetkey (command[i]);
          }
     }
}
definekey ("rn->enter_article_mode_then_do_something_else", ".", "group");
%}}}
static define move_cursor_to_next_group_with_unread_articles () %{{{
{
   % n -- move cursor to next group w/ unread articles. prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable starting_newsgroup = current_newsgroup ();

   do
     {
        !if (group_down_n (1))
          {
             % Leave group mode cursor on the starting_newsgroup and return.
             if (is_group_mode ())
               {
                  % We can be quick about it!
                  set_input_string (starting_newsgroup);
                  call ("add_group");
               }
             else
               {
                  % We have to be slow about it :-(
                  while (group_up_n (10000));    % skip to the top quickly.
                  while (current_newsgroup () != starting_newsgroup)
                    {
                       !if (group_down_n (1))
                         {
                            return ();           % unexpected error here!
                         }
                    }
               }
             return ();
          }
     }
   while (group_unread () == 0);
}
definekey ("rn->move_cursor_to_next_group_with_unread_articles", "n", "group");
%}}}
static define move_cursor_to_next_group () %{{{
{
   % N -- () = group_down_n (1); prompt again.
   () = group_down_n (1);

   group_mode_prompt ();
}
definekey ("rn->move_cursor_to_next_group", "N", "group");
%}}}
static define move_cursor_to_previous_group_with_unread_articles () %{{{
{
   % p -- move cursor to previous group w/ unread articles or stay where you
   %      are. prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable starting_newsgroup = current_newsgroup ();

   do
     {
        !if (group_up_n (1))
          {
             % Leave group mode cursor on the starting_newsgroup and return.
             if (is_group_mode ())
               {
                  % We can be quick about it!
                  set_input_string (starting_newsgroup);
                  call ("add_group");
               }
             else
               {
                  % We have to be slow about it :-(
                  while (group_up_n (10000));    % skip to the top quickly.
                  while (current_newsgroup () != starting_newsgroup)
                    {
                       !if (group_down_n (1))
                         {
                            return ();           % unexpected error here!
                         }
                    }
               }
             return ();
          }
     }
   while (group_unread () == 0);
}
definekey ("rn->move_cursor_to_previous_group_with_unread_articles", "p", "group");
%}}}
static define move_cursor_to_previous_group () %{{{
{
   () = group_up_n (1);

   group_mode_prompt ();
}
definekey ("rn->move_cursor_to_previous_group", "P", "group");
%}}}
static define move_cursor_to_first_group () %{{{
{
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   if (is_group_mode ())
     {
        call ("bob");
     }
   else
     {
        while (group_up_n (10000));
     }
}
definekey ("rn->move_cursor_to_first_group", "1", "group");
%}}}
static define move_cursor_to_topmost_group_with_unread_articles () %{{{
{
   % ^ -- move cursor to topmost group w/ unread articles. prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable starting_newsgroup = current_newsgroup ();
   move_cursor_to_first_group ();
   while (group_unread () == 0)
     {
        !if (group_down_n (1))
          {
             % Leave group mode cursor on the starting_newsgroup and return.
             if (is_group_mode ())
               {
                  % We can be quick about it!
                  set_input_string (starting_newsgroup);
                  call ("add_group");
               }
             else
               {
                  % We have to be slow about it :-(
                  while (group_up_n (10000));    % skip to the top quickly.
                  while (current_newsgroup () != starting_newsgroup)
                    {
                       !if (group_down_n (1))
                         {
                            return ();           % unexpected error here!
                         }
                    }
               }
             return ();
          }
     }
}
definekey ("rn->move_cursor_to_topmost_group_with_unread_articles", "^", "group");
%}}}
static define move_cursor_to_last_group () %{{{
{
   % $ -- leave the cursor on the last group. prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   if (is_group_mode ())
     {
        call ("eob");
     }
   else
     {
        while (group_down_n (10000));
     }
}
definekey ("rn->move_cursor_to_last_group", "$", "group");
%}}}
static define move_cursor_to_group () %{{{
{
   % g newsgroup -- goto named newsgroup.  prompt for subscription.  prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable requested_newsgroup = read_mini (sprintf ("******** %d unread articles in %s--read now? [ynq] g",
                                                      group_unread (), current_newsgroup ()),
                                             "", "");

   !if (strlen (requested_newsgroup)) return ();

   set_input_string (requested_newsgroup);
   call ("add_group");

   variable response = 0;
   if (get_group_flags () & GROUP_UNSUBSCRIBED)
     {
        message_now (sprintf ("Group %s is unsubscribed.  Do you want to subscribe? [yn]",
                              current_newsgroup ()), "", "");
        response = getkey ();
        if (orelse { response == 'y' } { response == 'Y' })
          {
             set_group_flags (get_group_flags () & ~GROUP_UNSUBSCRIBED);
          }
     }
}
definekey ("rn->move_cursor_to_group", "g", "group");
%}}}
static define move_cursor_forwards_to_regex_group () %{{{
{
   % /pattern -- find group matching pattern (including funky /r rule); wrap if
   %             necessary.  prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable
     starting_newsgroup = current_newsgroup (),
     allow_groups_with_0_unread_articles = 0,
     regex = read_mini (sprintf ("******** %d unread articles in %s--read now? [ynq] /",
                                 group_unread (), current_newsgroup ()),
                        "", "");

   if (regex == "") return ();
   if (string_match (regex, "/r$", 0))
     {
        regex = substr (regex, 1, strlen (regex) - 2);
        allow_groups_with_0_unread_articles = 1;
     }

   while (group_down_n (1))
     {
        if (string_match (current_newsgroup (), regex, 1))
          {
             if (orelse { group_unread () } { allow_groups_with_0_unread_articles })
               {
                  % Found a suitable group.
                  return ();
               }
          }
     }

   % Continue searching from the top by first placing the cursor at the top.
   if (is_group_mode ())
     {
        call ("bob");
     }
   else
     {
        while (group_up_n (10000));
     }

   % Then we search.
   while (group_down_n (1))
     {
        % No point in searching the same set of groups twice.  We're all done.
        % We found nothing that matches.
        if (current_newsgroup () == starting_newsgroup) return ();

        % It's a group we haven't examined before.
        if (string_match (current_newsgroup (), regex, 1))
          {
             if (orelse { group_unread () } { allow_groups_with_0_unread_articles })
               {
                  % Found a suitable group.
                  return ();
               }
          }
     }
}
definekey ("rn->move_cursor_forwards_to_regex_group", "/", "group");
%}}}
static define move_cursor_backwards_to_regex_group () %{{{
{
   % ?pattern -- find group matching pattern (including funky /r rule); wrap if
   %             necessary.  prompt again.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable
     starting_newsgroup = current_newsgroup (),
     allow_groups_with_0_unread_articles = 0,
     regex = read_mini (sprintf ("******** %d unread articles in %s--read now? [ynq] ?",
                                 group_unread (), current_newsgroup ()),
                        "", "");

   if (regex == "") return ();
   if (string_match (regex, "/r$", 0))
     {
        regex = substr (regex, 1, strlen (regex) - 2);
        allow_groups_with_0_unread_articles = 1;
     }

   while (group_up_n (1))
     {
        if (string_match (current_newsgroup (), regex, 1))
          {
             if (orelse { group_unread () } { allow_groups_with_0_unread_articles })
               {
                  % Found a suitable group.
                  return ();
               }
          }
     }

   % Continue searching from the bottom by first placing the cursor at the bottom.
   if (is_group_mode ())
     {
        call ("eob");
     }
   else
     {
        while (group_down_n (10000));
     }

   % Then we search.
   while (group_up_n (1))
     {
        % No point in searching the same set of groups twice.  We're all done.
        % We found nothing that matches.
        if (current_newsgroup () == starting_newsgroup) return ();

        % It's a group we haven't examined before.
        if (string_match (current_newsgroup (), regex, 1))
          {
             if (orelse { group_unread () } { allow_groups_with_0_unread_articles })
               {
                  % Found a suitable group.
                  return ();
               }
          }
     }
}
definekey ("rn->move_cursor_backwards_to_regex_group", "?", "group");
%}}}
static define unsubscribe_group () %{{{
{
   set_group_flags (get_group_flags () | GROUP_UNSUBSCRIBED);
   
   group_mode_prompt ();
}
definekey ("rn->unsubscribe_group", "u", "group");
%}}}
static define move_newsgroup () %{{{
{
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   variable group = read_mini (sprintf ("******** %d unread articles in %s--read now? [ynq] m",
                                        group_unread (), current_newsgroup ()),
                               "", "");

   set_input_string (group);
   call ("add_group");
   call ("move_group");
}
definekey ("rn->move_newsgroup", "m", "group");

%}}}
static define catchup () %{{{
{
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);

        group_mode_prompt ();
     }

   call ("catchup");
}
definekey ("rn->catchup", "c", "group");

%}}}

% Article and pager mode commands

static define article_mode_hook () %{{{
{
   % Set sorting method to unsorted (sort by article number) if -S is
   % not given.  If -S is given OR if subject search order was specified
   % (via pressing ^N), use threading and sorting by score and subject.
   if (orelse
         { substr (prefs["options"], "S", 0) }
         { subject_search_sort_order })
     {
        % Thread, then sort by score and subject.
        set_integer_variable ("sorting_method", 6);
     }
   else
     {
        % No sorting (article number sorting)
        set_integer_variable ("sorting_method", 0);
     }
   
   % Display the first page of the first article.  Then ask about
   % continuing.
   call ("article_page_down");
   article_mode_prompt ();
}

%}}}
static define article_mode_quit_hook () %{{{
{
   % Mark the articles in the delay_mark_current_article_unread_set
   % read then clear the set.

   variable message_ID = "";
   foreach (delay_mark_current_article_unread_set)
     {
        message_ID = ();
        if (message_ID == "") continue;
        () = locate_header_by_msgid (message_ID);
        set_header_flags (get_header_flags () | HEADER_READ);
     }
   delay_mark_current_article_unread_set = [""];
}

%}}}
static define show_next_unread_article () %{{{
{
   % Scan for next unread article and show that article.
   %
   % If used at the end of an article, the current article is
   %   not marked as read.  The current article is marked
   %   as read by the pager when it reads the last page of an
   %   article.
   
}

%}}}
static define restart_current_article () %{{{
{
   % Show the current article from the beginning.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
     }

   call ("article_bob");
}
%definekey ("rn->restart_current_article", "^R", "article");

%}}}
static define restart_current_article_verbosely () %{{{
{
   % Show the current article from the beginning with all the headers.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }

   if (headers_hidden_mode ()) call ("toggle_headers");
   restart_current_article ();   
}
%definekey ("rn->restart_current_article_verbosely", "v", "article");

%}}}
static define refresh_screen () %{{{
{
   % Refresh the screen.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }
   
   call ("refresh");
}
%definekey ("rn->refresh_screen", "^L", "article");

%}}}
static define restart_article_and_rot13 () %{{{
{
   % Show the current article from the beginning and ROT13 decode it.
   %
   % This macro will take a liberty with ROT13 encoding/decoding and
   % let the user toggle it on/off at will.  It's a lot easier to do
   % this with slrn and it makes more sense than the rn behavior is
   % described as being.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }

   call ("article_bob");
   call ("toggle_rot13");
}
%definekey ("rn->restart_article_and_rot13", "^X", "article");

%}}}
static define refresh_and_rot13 () %{{{
{
   % Refresh the screen then decrypt with rot13.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }
   
   call ("refresh");
   call ("toggle_rot13");   
}
%definekey ("rn->refresh_and_rot13", "X", "article");

%}}}
static define back_up_one_page () %{{{
{
   % Show the previous page of the article, if the article is showing at
   % all.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }

   if (_is_article_visible () & 1) { call ("article_page_up"); }
}
%definekey ("rn->back_up_one_page", "b", "article");

%}}}
static define quit_to_group_level () %{{{
{
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
     }

   call ("quit");
}
%definekey ("rn->quit_to_group_level", "q", "article");

%}}}
static define show_first_unread_article () %{{{
{
   % Go to the first unread article and display the first page of it.

   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }

   call ("header_bob");
   if (_is_article_visible () != 3) call ("article_line_down");
}
%definekey ("rn->show_first_unread_article", "^", "article");

%}}}
static define show_last_article () %{{{
{
   % Show the last article in the article list.  I don't know what
   % "actually, one past the last article" means so it will remain
   % unimplemented.
   
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
        article_mode_prompt ();
     }

   call ("header_eob");
   if (_is_article_visible () != 3) call ("article_line_down");
}
%definekey ("rn->show_last_article", "$", "article");

%}}}
static define enter_range_command () %{{{
{
   % Enter a command of the form "range{,range} command{:command}" and do
   % what that command asks.

   % In rn, the following commands are only used immediately after entering a
   % header number.  In this slrn macro set, a pseudo-command ":" will be
   % required before entering the "range{,range} command{:command}" syntax.
   % 
   % If slrn had a pre_header_number_hook() that executed *after* the new header
   % number was entered and *before* the cursor actually moved to the new header
   % number, we could do this just like rn does it.
   % 
   % The acceptable ranges are:
   %
   %    '.'        the current article index number
   %    number     the given article index number
   %    '$$'       the last article index number on the screen
   %    '$'        the last article
   %    p-q        where 'p' and 'q' are indices of the above form.
   %
   % The acceptable commands are:
   % 
   %    'm'        mark as unread
   %    'j'        mark as read
   %    's dest'   save to dest file
   %    '!command' shell escape
   %    '='        move the cursor to the lowest number in the range so the
   %               user can see the subject
   %    'C'        cancel

   variable input = read_mini (":", "", "");
   if (input == "") return ();

   !if (string_match (input, " ", 1)) error ("Syntax error: range{,range} command{:command}");

   variable
     % Break up the input into range(s) (first word) and command (all the rest).
     range = strchop (input, ' ', 0)[0],
     command = strchop (input, ' ', 0)[[1:]],

     % These are the numeric indices for the range(s) the user specified.
     range_list = Integer_Type[0],
     
     % These are temporary variables for the entries to be added to range_list, the range
     % being parsed, and the variables needed to parse within each range, respectively.
     subrange_entries = Integer_Type[0],
     subrange = "",
     offset = 0, len = 0,
     
     % The two ends of the range (for m-n range entries), and the increment (either +1 or -1).
     m = 0, n = 0, increment = 1;

   % The range should be in the form "range[,range[,range[,range[,range...]]]]"
   % 
   % Treat each range in turn.
   foreach (strchop (range, ',', 0))
     {
        % Each range will be of the following forms:
        %
        % "m" or "m-n" where m and n can be either numbers, ".", "$", or "$$".
        subrange = ();

        if (string_match (subrange, "\\([0-9.$]$?\\)-\\([0-9.$]$?\\)", 1))
          {
             % This is a range of articles of the form "m-n".

             % Get the "m" end of the range.
             (offset, len) = string_match_nth (1);
             m = substr (subrange, offset + 1, len);
             
             % Get the "n" end of the range.
             (offset, len) = string_match_nth (2);
             n = substr (subrange, offset + 1, len);

             % Handle shortcut values ("$", and ".") and numeric values.
             if (m == ".") m = header_cursor_pos ();
             else if (m == "$") m = SCREEN_HEIGHT - 3;
             else m = integer (m);
             if (n == ".") n = header_cursor_pos ();
             else if (n == "$") n = SCREEN_HEIGHT - 3;
             else n = integer (n);

             % Make the between values.
             if (andelse { m != NULL } { n != NULL } { m > n }) increment = -1;
             subrange_entries = [[m:n:increment]];

             % Add them to the list of articles to visit.
             range_list = [range_list, subrange_entries];
          }
        else if (string_match (subrange, "\\([0-9.$]$?\\)", 1))
          {
             % This is a single article of the form "m".

             % Get the "m" value (there is no "n" end because there is no range).
             (offset, len) = string_match_nth (1);
             m = integer (substr (subrange, offset + 1, len));

             % Add that value to the list of articles to visit.
             range_list = [range_list, m];
          }
        else error (sprintf ("Syntax error: [%s] article number must be a single value (m) or a range of values (m-n).", subrange));
     }
   
   % Parse the command(s) to perform.
   %
   % The acceptable commands are:
   % 
   %    'm'        mark as unread
   %    'j'        mark as read
   %    's dest'   save to dest file
   %    '!command' shell escape
   %    '='        move the cursor to the lowest number in the range so the
   %               user can see the subject
   %    'C'        cancel

   % FIX ME: can't handle "js blah" which should mark the range(s) read and save them to
   % "blah". can't handle the admittedly silly "jmjmjm" which toggles the articles read
   % then unread 3 times.
   %
   % This is being "fixed" by introducing a command separator character ";".  So this is
   % the new way to do the above: "j;s blah" and "j;m;j;m;j;m".  This lets us handle saving
   % and other commands after saving.  "\;" will mean a literal semicolon (for saving and
   % shell escapes).

   variable
     % Article being worked on.
     article_to_jump_to = 0,
     
     % Command being worked on.
     current_command = "",
     
     % Path to save article to/shell command/etc.
     argument = "",
     
     % filehandle for saving articles.
     fp = 0;

   foreach (range_list)
     {
        article_to_jump_to = ();
        !if (jump_to_article (article_to_jump_to)) continue;

        % Split up the command input string into individual commands,
        % separated or terminated by ";".
        foreach (strchop (command, ';', '\\'))
          {
             % Clean up the command being worked on.  If the command being
             % worked on is the empty string, skip to the next command.
             current_command = strtrim ( () );
             if (current_command == "") continue;

             % Do the command for the current article.
             switch (current_command[0])
               { orelse { case ('s') } { case ('S') } :
                    {
                       % Save to the indicated path.
                       if (current_command[1] != ' ') argument = current_command[[1:]];
                       else argument = strchop (current_command, ' ', 0)[1];

                       % Save article to file.
                       fp = fopen (argument, "a");
                       if (fp == NULL) verror ("Could not open %s for writing.", argument);
                       if (-1 == fputs ("\n" + article_as_string (), fp))
                         verror ("Could not write to %s: %s", argument, errno_string (errno));
                       if (-1 == fclose (fp))
                         verror ("Could not close %s: %s", argument, errno_string (errno));
                    }
               }
               { case ('!') :
                    {
                       % Shell out with the indicated shell command.
                       if (current_command[1] != ' ') argument = current_command[[1:]];
                       else argument = strchop (current_command, ' ', 0)[1];

                       () = system (argument);
                    }
               }
               { case ('m') :
                    {
                       % Mark article as unread.
                       set_header_flags (get_header_flags () & ~HEADER_READ);
                    }
               }
               { case ('j') :
                    {
                       % Mark article as read.
                       set_header_flags (get_header_flags () | HEADER_READ);
                    }
               }
               { case ('!') :
                    {
                       % Move the cursor to the lowest number in the range_list
                       % so the user can see the subject.
                       
                    }
               }
               { case ('C') :
                    {
                       % Cancel the articles in range_list.  How to do this so it
                       % is sane and like rn?
                       
                    }
               }
               { error (sprintf ("Syntax error: [%s] bad command.", command)); }
          }
     }
}
%definekey ("rn->enter_range_command", ":", "article");

%}}}
static define read_next_page_or_scan_forward_for_next_unread_article () %{{{
{
   % Display the next page of an article or
   % scan forward for the next unread article.
   
   switch (get_next_art_pgdn_action ())
     % Next page of the article will be shown.
     { case (0) : call ("article_page_down"); }
     % The next article will be shown.
     { case (1) : show_next_unread_article (); }
     % The next newsgroup will be entered.
     { case (2) : call ("quit"); }
}
%definekey ("rn->read_next_page_or_scan_forward_for_next_unread_article", " ", "article");

%}}}
static define mark_current_article_unread () %{{{
{
   % Mark the current article unread.
   set_header_flags (get_header_flags () & ~HEADER_READ);
}
%definekey ("rn->mark_current_article_unread", "m", "article");

%}}}
static define command_delay_mark_as_unread () %{{{
{
   % Mark the current article unread on group exit.  Until then, mark
   % the article read.
   %
   % To do this, we will add the article's message-ID to a list and
   % have a article_mode_quit_hook() do the work of marking the articles
   % as read and clearing the list.
   
   variable message_ID = extract_article_header ("Message-ID");
   
   % In case where() tanks completely, we'll keep on running.  We reset the
   % ERROR_BLOCK afterwards to let other errors occur normally.  This is as
   % close to throw-catch as S-Lang gets right now.
   ERROR_BLOCK { _clear_error (); }
   if (where (delay_mark_current_article_unread_set == message_ID)[0]) return ();
   ERROR_BLOCK {}

   % Can't find the current message-ID, so we add it. 
   delay_mark_current_article_unread_set = [delay_mark_current_article_unread_set, message_ID];
   delay_mark_current_article_unread_set = reshape (delay_mark_current_article_unread_set, [length (delay_mark_current_article_unread_set)]);
}
command_map["M"] = &command_delay_mark_as_unread;
%definekey ("rn->delay_mark_current_article_unread", "M", "article");

%}}}
static define command_mark_as_unread () %{{{
{
   set_header_flags (get_header_flags () | HEADER_READ);
}
command_map["m"] = &command_mark_as_unread;
%}}}
static define command_mark_as_read () %{{{
{
   set_header_flags (get_header_flags () | HEADER_READ);
}
command_map["j"] = &command_mark_as_read;

%}}}
static define command_save_article (destination) %{{{
{
   
}
command_map["s"] = &command_save_article;

%}}}
static define command_shell_out (command) %{{{
{
   () = system (command);
}
command_map["!"] = &command_shell_out;

%}}}
static define command_cancel_article () %{{{
{
   % Preserve the prefix argument and leave it available for use inside
   % this macro.  Restore the prefix argument when this macro returns.
   variable prefix_argument = get_prefix_arg ();
   reset_prefix_arg ();

   EXIT_BLOCK
     {
        if (prefix_argument !=-1) set_prefix_argument (prefix_argument);
     }
   
   call ("cancel");
}
command_map["C"] = &command_cancel_article;

%}}}
static define junk_current_article () %{{{
{
   % Mark the current article as read and skip to the end of the
   % article (if seeing an article).
   command_mark_as_read ();

   if (_is_article_visible () & 1) call ("article_eob");
}
%definekey ("rn->junk_current_article", "j", "article");

%}}}

{
   % Scan forward for article containing given pattern in the subject,
   % headers, or anywhere in the article.  The overall syntax for the
   % search command is:
   %
   %     /pattern[/modifiers[:command...]]
   %
   % The modifiers are:
   % 
   %      h       Scan forward for article containing pattern in the header.
   %      a       Scan forward for article containing pattern anywhere in
   %              article.
   %      r       Scan read articles also.  Searches only scan unread articles
   %              by default.
   %      c       Make search case sensitive.  Searches are case-insensitive
   %              by default.
   %
   % So commands can be tied to searches (which means the commands are
   % executed immediately after finding each matching article).  The
   % commands are:
   %
   %     m       mark as unread
   %     M       delayed mark as unread
   %     j       mark as read
   %     s dest  save the article to dest (and "|dest" means pipe to dest)
   %     !cmd    shell out and run cmd
   %     =       print the subject (might not be implementable)
   %     C       cancel the article
   %
   % The ? is the backwards search command, it works similarly.

   variable
     % Get the input from the user.
     input = read_mini ("/", "", ""),
   
     % A temp string for processing the input string.
     processing = "",

     % The regular expression pattern to look for ("\/" is required to escape
     % the "/" char in the user's input).
     pattern = "",
     
     % The modifiers block.
     modifiers = "",
     
     % The set of commands (one command per array cell).
     commands = String_Type[0],
     
     % Set up bits to indicate what the search parameters are:
     %
     % What is the scope of the search (subject | headers | whole-article)
     search_scope = "subject",
     % Are we searching read articles in addition to unread articles?
     search_read_articles_too = 0,
     % Are we searching in a case-sensitive way?
     search_is_case_sensitive = 0,

     % This holds the string being matched against for the search pattern.
     scope = "",
     
     % This is a temporary variable to hold the article.
     article = "";

   if (input == "") return ();

   % Scanning from left to right, if there's an unescaped slash, it separates
   % the pattern from the rest.
   if (string_match (input, "[^\\\\]/", 1))
     {
        processing = strchop (input, '/', '\\');
        pattern = processing[0];
        processing = strjoin (processing[[1:]], "/");

        % The modifiers block is everything right of the pattern block
        % through the leftmost unescaped colon.
        modifiers = strchop (processing, ':', '\\')[0];
        commands = strchop (processing, ':', '\\')[[1:]];
     }
   else pattern = input;

   % Parse the modifiers and set the search_* bits and state to correct values.
   if (string_match (modifiers, "h", 1) and string_match (modifiers, "a", 1))
     error ("Incorrect syntax: only one of \'h\' and \'a\' are allowed.");
   if (string_match (modifiers, "h", 1)) search_scope = "headers";
   else if (string_match (modifiers, "a", 1)) search_scope = "whole-article";
   if (string_match (modifiers, "r", 1)) search_read_articles_too = 1;
   if (string_match (modifiers, "c", 1)) search_is_case_sensitive = 1;

   % Search the rest of the article list from the article immediately below
   % the cursor through to the end of the article list.
   %
   % First we search the current thread for matching articles.

   variable
     % The starting thread was collapsed.  This is used to restore the
     % state of the starting thread.
     starting_thread_was_collapsed = is_thread_collapsed (),

     % The article we began this search with.
     starting_article = extract_article_header ("Message-ID"),

     % The current thread was collapsed.  This is used to restore the
     % state of the currently examined thread.
     thread_was_collapsed = 0,

     % A temp. variable and a variable to hold the size of the thread.
     i = 0, t_size = 0;
   
   % This will be handled in two parts.  The first part handles the thread
   % the cursor starts on.  First we have to position the cursor at the
   % proper article to begin searching.
   if (is_thread_collapsed ())
     {
        % We're at the top of a thread.
        uncollapse_thread ();
        thread_was_collapsed = 1;
        () = header_down (1);
     }
   else
     {
        % We're somewhere within a thread, possibly in a one-article thread.
        if (whole_thread_size () > 1)
          {
             % There are other articles in the thread below this article.
             thread_was_collapsed = 1;
             () = header_down (1);
          }
        else
          {
             % There are no other articles in the thread below this article.
             % So we begin searching at the start of a new thread.
             () = header_down (1);
          }
     }

   % Now the cursor is at the first article that needs to be scanned for
   % matching.
   t_size = whole_thread_size ();
   for (i = 1; i <= t_size; i++)
     {
        if (found_a_match (search_read_articles_too, search_is_case_sensitive, search_scope, pattern)) return ();
        if (i != t_size) () = header_down (1);
     }
   % So we didn't find a match.  Time to do some clean up work on the starting
   % thread.
   if (thread_was_collapsed) collapse_thread ();
   !if (header_down (1))
     {
        % There are no more articles to search.  The search has failed.
        () = locate_header_by_msgid (starting_article, 0);
        message ("No matching articles found.");
        return ();
     }

   % The second part handles every other thread below the starting thread.
   do
     {
        t_size = whole_thread_size ();
        thread_was_collapsed = is_thread_collapsed ();
        for (i = 1; i <= t_size; i++)
          {
             
          }
        if (thread_was_collapsed) collapse_thread ();
     }
   while (header_down (1)); % next visible thread.

   if (starting_thread_was_collapsed) uncollapse_thread ();
   !if (header_down (1)) error ("No more articles to search.");
   do
     {
        t_size = whole_thread_size ();
        thread_was_collapsed = is_thread_collapsed ();
        for (i = 1; i <= t_size; i++)
          {
             if (found_a_match (search_read_articles_too, search_is_case_sensitive, search_scope, pattern)) return ();
             if (andelse { i == 1 } { thread_was_collapsed }) uncollapse_thread ();
             if (i != t_size) () = header_down (1);
          }
        if (thread_was_collapsed) collapse_thread ();
     }
   while (header_down (1));

   % We did not find anything, so we'll put the cursor back and tell the user.
   () = locate_header_by_msgid (starting_article, 0);
   if (starting_thread_was_collapsed) collapse_thread ();
   message ("No matching articles found.");
}
definekey ("Global->search2", "/", "article");

#iffalse % Ignore the macros behind the curtain... %{{{
static define range_minus () %{{{
{
   % Start a range command with a minus.  This range means a list of commands
   % applied to a contiguous series of articles.

   !if (is_a_range_request) { return (); }

   variable
     % The command the user supplies.
     command = read_mini (sprintf ("%d-", get_header_number ()),
                          "", ""),
   
     % The offset and length used to parse the command.
     offset = 0, len = 0,
     
     % The parts of the command.
     end_range = 0, action = "", argument = "";

   !if (strlen (command)) { return (); }
   
   % Either the command has a dot for the other end of the range or it has a
   % real number.  Let's figure out which.  First the real number.
   if (string_match (command, "\\([0-9]+\\) *\\([mjs!=C]\\)\\(.*\\)", 1))
     {
        % The user gave a number for the end of the range.
        
        % Parse out the end of the range...
        (offset, len) = string_match_nth (1);
        end_range = integer (substr (command, offset + 1, len));
        
        % ...the action to take...
        (offset, len) = string_match_nth (2);
        action = substr (command, offset + 1, len);
        
        % ...and the argument to that action.
        (offset, len) = string_match_nth (3);
        argument = substr (command, offset + 1, len);
     }
   else if (string_match (command, ". *\\([mjs!=C]\\)\\(.*\\)", 1))
     {
        % The user inferred a number for the end of the range.  The inference
        % specified the current article's header number.
        end_range = header_cursor_pos ();

        % ...the action to take...
        (offset, len) = string_match_nth (1);
        action = substr (command, offset + 1, len);

        % ...and the argument to that action.
        (offset, len) = string_match_nth (2);
        argument = substr (command, offset + 1, len);
     }
   else error ("Syntax error in command.");

   % Now end_range, action, and argument have valid values.  Let's do
   % something useful with them.

   if (action == "!")
     {
        % Shell out.
        () = system (argument);
     }
%   else if (action == "=")
%     {
%        % Move the cursor to the low end of the range and display
%        % that article.
%        if (get_article_index_number () < end_range)
%          {
%             % Move cursor to the article at end_range.
%             
%          }
%        
%        restart_current_article ();
%     }
   else
     {
        switch (action)
          { case ('m') : 
          }
     }
   
   is_a_range_request = 0;
}
%definekey ("rn->range_minus", "-", "article");

%}}}
%}}}
